//
//  Win32NativeUiInterface.h
//  scummvm
//
//
//

#ifndef Win32NativeUiInterface_h
#define Win32NativeUiInterface_h

#include "enhanced/ui/NativeUiInterface.h"
#include "dists/msvc15/netwrapper/UnmanagedWrapper.h"
#include "common/list.h"
#include "common/str.h"

using namespace NetWrapper;

#if defined(APP_STORE_GOG)
	class AuthListener;
	class UserStatsAndAchievementsRetrieveListener;
	class StatsAndAchievementsStoreListener;
#endif


class Win32NativeUiInterface : public Enhanced::NativeUiInterface {

private:
	UnmanagedWrapper* _wrapper;
	vector<const wchar_t*>* _localizedStrings;
	//bool _backToMenu;
	//bool _isLoadingGameFromMainMenu;
	//bool _isStartingNewGame;
	bool _showingMenu;
	bool _secondInstance;
	HANDLE _mutexHandle;
	//int _level;

	void ShowMenuInternal(int menuAction);
	void ShowSettingsSubMenu(int menuAction, SettingSubMenuType type);
	void LoadStrings();
	wchar_t* CreateWCharString(wstring string);
	void DeleteWCharVector(vector<const wchar_t*>* wcharVector);
	vector<const wchar_t*>* WStringToWCharVector(vector<wstring>& original);
	ConvertedSetting* CreateSetting(const BaseSetting* setting);
	void DeleteSetting(ConvertedSetting* setting);
	int GetStates();
	void MaximizeGameWindow();
	void MinimizeGameWindow();
	void BeforeDialog();
	void AfterDialog();
public:

	Win32NativeUiInterface();
	~Win32NativeUiInterface();

	/**
	 * Show the game menu.
	 * If showSplash is true, start by showing the game's splash screen transition.
	 */
	virtual void showMenu(bool showSplash);

	/**
	 * Hide the game menu if it's displayed.
	 */
	virtual void hideMenu();

	/**
	 * Show the game's tutorial video. (same as a click on 'Tutorial')
	 *
	 * This function is exposed in the interface, to allow the tutorial to be started programatically by the logic.
	 * For example, when prompting for the tutorial in the first game run.
	 */
	virtual void showTutorialVideo();


	/**
	 * This function must be implemented by the UI, to allow for enabled\disabled state changes in the settings.
	 *
	 * This function will only be called after a "setValue" operation on a setting.
	 *
	 * When this function is called, the UI should call isEnabled() on the currently displayed settings, and refresh their displayed status as needed.
	 *
	 *  Example (for testing this functionality): Select "subtitles only", and the "voice" volume slider should become disabled.
	 */
	virtual void refreshSettingStates();


	//
	// The following functions enable showing interactive dialogs, and receiving results.
	// A dialog might be shown during the game, or on top of the game menu.
	// Should be implemented by native UI.
	//

	/**
	 * Show a dialog with a message, and 1 to 3 buttons.
	 * The dialog should be shown according to each game's skin.
	 * When the user presses a button, the dialog should be dismissed and the listener should be called with the result.
	 *
	 * Examples of possible dialog types from the existing Android implementation (only for a visual reference):
	 * 1 button - whenever the language setting is changed and the user returns to the main menu.
	 * 2 buttons - whenever the game is active and the user presses 'Quit' in the main menu.
	 * 3 buttons - "Appirater" dialog - when the game prompts for rating on the app store.
	 *
	 * SPECIAL CASE: "The 7th Guest" shows 2-button dialogs as a full screen sphinx dialog. For example, when quitting.
	 */
	virtual void showAlertDialog(wstring message, int buttonNum, wstring button1, wstring button2, wstring button3, shared_ptr<Enhanced::AlertDialogListener> listener);

	/**
	 * Show a dialog with a message, a text input field containing the initial input, and 2 buttons.
	 * The dialog should be shown according to each game's skin.
	 * When the user presses a button, the dialog should be dismissed and the listener should be called with the result input (or no result).
	 *
	 * Example of a possible dialog type from the existing Android implementation (only for a visual reference):
	 * Whenever the user pressed "Save" in the main menu and chooses a save slot.
	 */
	virtual void showInputDialog(wstring message, string initialInput, wstring negativeButton, wstring positiveButton, shared_ptr<Enhanced::InputDialogListener> listener);


	/**
	 * Show a dialog with a message, a list containing choosable items, and 1 or 2 buttons.
	 * The dialog should be shown according to each game's skin.
	 * When the user presses an item, the dialog should be dismissed and the listener should be called with the result selection.
	 * When the user presses the negative button, the dialog should be dismissed the the listener should be called with no result.
	 *
	 * Example of possible dialog types from the existing Android implementation (only for a visual reference):
	 * Whenever the user pressed "Save" or "Load" in the main menu.
	 *
	 * SPECIAL CASE: "I Have No Mouth And I Must Scream" shows list dialogs with 2 buttons (negative and positive), unlike other games.
	 * In this game, the dialog should be dismissed and the listener should be called only when one of the buttons is pressed - not the list items.
	 */
	virtual void showListDialog(wstring message, vector<wstring> items, int buttonNum, wstring negativeButton, wstring positiveButton, shared_ptr<Enhanced::ListDialogListener> listener);

	/**
	 * Show a modal progress dialog with a message.
	 *
	 * Example: When saving or loading a game, the "please wait..." dialog appears.
	 */
	virtual void showProgressDialog (wstring message);

	/**
	 * Hide the current progress dialog if it's displayed.
	 */
	virtual void hideCurrentDialog();








	//
	// Misc. functions to be implemented by the native UI:
	//

	/**
	 * Open the relevant app store page for rating the app (the result of "Appirater").
	 */
	virtual void openAppStoreForRating();

	/**
	 * Return the current OS locale, to be used for accessing UI strings.
	 */
	virtual Enhanced::Locale getLocale();

	/**
	 * Show the provided text at the game's subtitle view.
	 */
	virtual void showSubtitles(wstring text);

	/**
	 * Hide the game's subtitle view.
	 */
	virtual void hideSubtitles();

	/**
	 * Show the static, full-screen developer logo.
	 * In all of MojoTouch's games, the logo is the MojoTouch logo.
	 * In IHNMAIMS, the logo is DotEmu's logo.
	 */
	virtual void showDeveloperLogo1();

	/**
	 * "The 7th Guest" - show the Trilobyte logo.
	 */
	virtual void showDeveloperLogo2();

	/**
	 * Hide the current developer logo if it's displayed.
	 */
	virtual void hideDeveloperLogo();

	/**
	 * Send game start statistics containing the settings used for starting the game.
	 */
	virtual void sendGameStartAnalytics(SettingInfo& settingInfo);

	virtual void onAchievementReached(int achievementId);

	virtual void onSaveConfigSuccessful() {}

	virtual void onLoadingConfig() {}

	virtual void onSaveSuccessful(int slot) {}

	virtual void onLoading(int slot) {}

	virtual bool useTouchScreen();

	virtual bool isSmallScreen();

	virtual bool useHWMouse();

	virtual bool useGameController();

	virtual void tick();


#if defined(APP_STORE_GOG)
private:
	AuthListener *_authListener;
	UserStatsAndAchievementsRetrieveListener *_userStatsAndAchievementsRetrieveListener;
	StatsAndAchievementsStoreListener *_statsAndAchievementsStoreListener;

	void GOGSetAchievement(const char* ID);

	typedef Common::List<Common::String> AchievementsList;
	AchievementsList _achivementsQueue;

public:
	int _gogAPIReady;

	void processAchievementsQueue();
#endif

};

#endif /* Win32NativeUiInterface_h */
