/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#include "common/debug.h"
#include "common/file.h"
#include "common/system.h"
#include "common/ustr.h"

#include "graphics/fonts/ttf.h"
#include "graphics/font.h"
#include "graphics/fontman.h"
#include "graphics/surface.h"

#include "video/subtitles.h"

#include "enhanced/GameSettingsManager.h"

namespace Video {

SRTParser::SRTParser() {
}

SRTParser::~SRTParser() {
	cleanup();
}

void SRTParser::cleanup() {
	for (Common::Array<SRTEntry *>::const_iterator item = _entries.begin(); item != _entries.end(); ++item)
		delete *item;

	_entries.clear();
}

bool parseTime(const char **pptr, uint32 *res) {
	int hours, mins, secs, msecs;
	const char *ptr = *pptr;

	hours = (*ptr++ - '0') * 10;
	hours += *ptr++ - '0';

	if (hours > 24 || hours < 0)
		return false;

	if (*ptr++ != ':')
		return false;

	mins = (*ptr++ - '0') * 10;
	mins += *ptr++ - '0';

	if (mins > 60 || mins < 0)
		return false;

	if (*ptr++ != ':')
		return false;

	secs = (*ptr++ - '0') * 10;
	secs += *ptr++ - '0';

	if (secs > 60 || secs < 0)
		return false;

	if (*ptr++ != ',')
		return false;

	msecs =  (*ptr++ - '0') * 100;
	msecs += (*ptr++ - '0') * 10;
	msecs +=  *ptr++ - '0';

	if (msecs > 1000 || msecs < 0)
		return false;

	*res = 1000 * (3600 * hours + 60 * mins + secs) + msecs;

	*pptr = ptr;

	return true;
}

int SRTEntryComparator(const void *item1, const void *item2) {
	const SRTEntry *l = *(const SRTEntry * const *)item1;
	const SRTEntry *r = *(const SRTEntry * const *)item2;

	return l->start - r->start;
}

int SRTEntryComparatorBSearch(const void *item1, const void *item2) {
	const SRTEntry *k = *(const SRTEntry * const *)item1;
	const SRTEntry *i = *(const SRTEntry * const *)item2;

	if (k->start < i->start)
		return -1;

	if (k->start > i->end - 1)
		return 1;

	return 0;
}

bool SRTParser::parseFile(const char *fname) {
	Common::File f;

	cleanup();

	if (!f.open(fname)) {
		return false;
	}

	byte buf[3];
	f.read(buf, 3);

	int line = 0;

	// Skip UTF header if present
	if (buf[0] != 0xef || buf[1] != 0xbb || buf[2] != 0xbf)
		f.seek(0);

	while (!f.eos()) {
		Common::String sseq, stimespec, stmp, text;

		sseq = f.readLine(); line++;
		stimespec = f.readLine(); line++;
		text = f.readLine(); line++;

		if (sseq.empty()) {
			if (f.eos()) {
				// Normal end of stream
				break;
			} else {
				warning("Bad SRT file format (spec): %s at line %d", fname, line);
				break;
			}
		}

		if (stimespec.empty() || text.empty()) {
			warning("Bad SRT file format (spec): %s at line %d", fname, line);
			break;
		}

		// Read all multiline text
		while (!f.eos()) {
			stmp = f.readLine(); line++;

			if (!stmp.empty()) {
				text += '\n';
				text += stmp;
			} else {
				break;
			}
		}

		uint32 seq = atol(sseq.c_str());
		if (seq == 0) {
			warning("Bad SRT file format (seq): %s at line %d", fname, line);
			break;
		}

		// 00:20:41,150 --> 00:20:45,109
		if (stimespec.size() < 29) {
			warning("Bad SRT file format (timespec length %d): %s at line %d", stimespec.size(), fname, line);
			break;
		}

		const char *ptr = stimespec.c_str();
		uint32 start, end;
		if (!parseTime(&ptr, &start)) {
			warning("Bad SRT file format (timespec start): %s at line %d", fname, line);
			break;
		}

		while (*ptr == ' ')
			ptr++;

		while (*ptr == '-')
			ptr++;

		if (*ptr != '>') {
			warning("Bad SRT file format (timespec middle ('%c')): %s at line %d", *ptr, fname, line);
			break;
		}

		ptr++;

		while (*ptr == ' ')
			ptr++;

		if (!parseTime(&ptr, &end)) {
			warning("Bad SRT file format (timespec end): %s at line %d", fname, line);
			break;
		}

		_entries.push_back(new SRTEntry(seq, start, end, text));
	}

	qsort(_entries.data(), _entries.size(), sizeof(SRTEntry *), &SRTEntryComparator);

	debug(6, "SRTParser: Loaded %d entries", _entries.size());

	return true;
}

Common::String SRTParser::getSubtitle(uint32 timestamp, uint32 *duration) {
	SRTEntry  test(0, timestamp, 0, "");
	SRTEntry *testptr = &test;
	if (duration)
		*duration = 0;

	const SRTEntry **entry = (const SRTEntry **)bsearch(&testptr, _entries.data(), _entries.size(), sizeof(SRTEntry *), &SRTEntryComparatorBSearch);

	if (entry == NULL)
		return "";

	if (duration)
		*duration = (*entry)->end - (*entry)->start;

	return (*entry)->text;
}

Subtitles::Subtitles() : _loaded(false), _font(nullptr), _prevExpiry(0) {
	_surface = new Graphics::Surface();
}

Subtitles::~Subtitles() {
	delete _surface;
}


// HACK. This must be unified and moved to common/ FIXME
static Common::U32String convertUtf8ToUtf32(const Common::String &str) {
	// The String class, and therefore the Font class as well, assume one
	// character is one byte, but in this case it's actually an UTF-8
	// string with up to 4 bytes per character. To work around this,
	// convert it to an U32String before drawing it, because our Font class
	// can handle that.
	Common::U32String u32str;
	uint i = 0;
	while (i < str.size()) {
		uint32 chr = 0;
		if ((str[i] & 0xF8) == 0xF0) {
			chr |= (str[i++] & 0x07) << 18;
			chr |= (str[i++] & 0x3F) << 12;
			chr |= (str[i++] & 0x3F) << 6;
			chr |= (str[i++] & 0x3F);
		} else if ((str[i] & 0xF0) == 0xE0) {
			chr |= (str[i++] & 0x0F) << 12;
			chr |= (str[i++] & 0x3F) << 6;
			chr |= (str[i++] & 0x3F);
		} else if ((str[i] & 0xE0) == 0xC0) {
			chr |= (str[i++] & 0x1F) << 6;
			chr |= (str[i++] & 0x3F);
		} else {
			chr = (str[i++] & 0x7F);
		}
		u32str += chr;
	}
	return u32str;
}

#ifdef USE_FREETYPE2
    
void Subtitles::setFont(const char *fontname, int height) {
	Common::File file;

	//_fontHeight = height;
    _fontHeight = g_system->getOverlayWidth() / 33;

	if (file.open(fontname)) {
		_font = Graphics::loadTTFFont(file, _fontHeight, Graphics::kTTFSizeModeCell);

        LOGD("Font Loaded: %s", file.getName());

	}

	if (!_font) {
		warning("Cannot load font %s directly", fontname);
		_font = FontMan.getFontByName(fontname);
	}

	if (!_font) {
		warning("Cannot load font %s", fontname);

		_font = FontMan.getFontByUsage(Graphics::FontManager::kBigGUIFont);
	}

}
    
#endif

void Subtitles::loadSRTFile(const char *fname) {
	debug(1, "loadSRTFile('%s')", fname);

	_loaded = _srtParser.parseFile(fname);
}

void Subtitles::setBBox(const Common::Rect bbox) {
	_bbox = bbox;

    //LOGD("Subtitles::setBBox Height: %d", _bbox.height());
    //LOGD("Subtitles::setBBox Width: %d", _bbox.width());
    //LOGD("Subtitles::setBBox Left: %d", _bbox.left);
    //LOGD("Subtitles::setBBox Top: %d", _bbox.top);
    //LOGD("Subtitles::setBBox Bottom: %d", _bbox.bottom);
    
    SHADOW = g_system->getOverlayHeight() / 500;
	_surface->create(_bbox.width() + SHADOW * 2, _bbox.height() + SHADOW * 2, g_system->getOverlayFormat());
}

void Subtitles::setColor(byte r, byte g, byte b) {
	_color = _surface->format.ARGBToColor(255, r, g, b);
	_blackColor = _surface->format.ARGBToColor(255, 0, 0, 0);
	_transparentColor = _surface->format.ARGBToColor(0, 0, 0, 0);
}

void Subtitles::drawSubtitle(uint32 timestamp, bool force) {
	uint32 duration;
	uint32 currTime = g_system->getMillis();

	if (!_loaded)
		return;

	Common::String subtitle = _srtParser.getSubtitle(timestamp, &duration);

	if (!force && _prevExpiry && currTime < _prevExpiry)
		return;

	update();

	if (!force && subtitle == _prevSubtitle)
		// Not yet reached next one
		return;

	// Save subtitle details
	_prevSubtitle = subtitle;
	_prevExpiry = currTime + duration;

#ifdef DEBUG
	debug(1, "%d: %s", timestamp, subtitle.c_str());
	LOGD("Subtitles::drawSubtitle: %d: %s", timestamp, subtitle.c_str());
    //LOGD("Subtitles::drawSubtitle Height: %d", _surface->h);
    //LOGD("Subtitles::drawSubtitle Width: %d", _surface->w);
#endif
	_surface->fillRect(Common::Rect(0, 0, _surface->w, _surface->h), _transparentColor);

	Common::Array<Common::U32String> lines;

    _font->wordWrapText(convertUtf8ToUtf32(subtitle), _surface->w, lines);
 
    //LOGD("Subtitles::Bbox Height: %d", _bbox.height());
    //LOGD("Subtitles::Bbox Width: %d", _bbox.width());
    //LOGD("Subtitles::Bbox Left: %d", _bbox.left);
    //LOGD("Subtitles::Bbox Top: %d", _bbox.top);
    //LOGD("Subtitles::Bbox Bottom: %d", _bbox.bottom);
    
    int y = 0;
    for (int i = 0; i < lines.size(); i++) {
        if (Enhanced::GameSettingsManager::instance()->getStringValue(SETTING_KEY_LANGUAGE).contains(SETTING_VALUE_LANGUAGE_HE)) {
            std::reverse(lines[i].begin(), lines[i].end());
        }
        if (lines.size() == 1) {
            y += _font->getFontHeight() / 1.5;
        }
        
        _font->drawString(_surface, lines[i], 0, y, _surface->w, _blackColor, Graphics::kTextAlignCenter);
        _font->drawString(_surface, lines[i], SHADOW * 2, y, _surface->w, _blackColor, Graphics::kTextAlignCenter);
        _font->drawString(_surface, lines[i], 0, y + SHADOW * 2, _surface->w, _blackColor, Graphics::kTextAlignCenter);
        _font->drawString(_surface, lines[i], SHADOW * 2, y + SHADOW * 2, _surface->w, _blackColor, Graphics::kTextAlignCenter);

        _font->drawString(_surface, lines[i], SHADOW, y + SHADOW, _surface->w, _color, Graphics::kTextAlignCenter);

        y += _font->getFontHeight();

        if (y > _bbox.bottom)
            break;
    }

	g_system->copyRectToOverlay(_surface->getPixels(), _surface->pitch, _bbox.left, _bbox.top, _bbox.width(), _bbox.height());
	g_system->updateScreen();
}

void Subtitles::update() {
	uint32 currTime = g_system->getMillis();

	if (_prevExpiry && currTime >= _prevExpiry) {
		// Finished current subtitle
		_prevExpiry = 0;
		g_system->clearOverlay();
	}
}

} // End of namespace Video
